@@ -2,6 +2,7 @@ import fileinput
 import logging
 import os
 from dataclasses import dataclass
+from functools import wraps
 from os.path import dirname, exists, join
 from typing import List, cast
 
@@ -10,12 +11,13 @@ import sentry_sdk
 from dotenv import load_dotenv
 from fcache.cache import FileCache  # type: ignore
 from sentry_sdk.integrations.aiohttp import AioHttpIntegration
+from sentry_sdk.integrations.logging import LoggingIntegration
 
 from dipdup import __spec_version__, __version__, spec_reindex_mapping, spec_version_mapping
 from dipdup.codegen import DEFAULT_DOCKER_ENV_FILE, DEFAULT_DOCKER_IMAGE, DEFAULT_DOCKER_TAG, DipDupCodeGenerator
 from dipdup.config import DipDupConfig, LoggingConfig, PostgresDatabaseConfig
 from dipdup.dipdup import DipDup
-from dipdup.exceptions import ConfigurationError, MigrationRequiredError
+from dipdup.exceptions import ConfigurationError, DipDupError, MigrationRequiredError
 from dipdup.hasura import HasuraGateway
 from dipdup.utils import set_decimal_context, tortoise_wrapper
 
@@ -29,12 +31,52 @@ class CLIContext:
     logging_config: LoggingConfig
 
 
+def cli_wrapper(fn):
+    @wraps(fn)
+    async def wrapper(*args, **kwargs):
+        try:
+            return await fn(*args, **kwargs)
+        except KeyboardInterrupt:
+            pass
+        except DipDupError as e:
+            _logger.critical(e.__repr__())
+            _logger.info(e.format())
+            quit(e.exit_code)
+
+    return wrapper
+
+
+def init_sentry(config: DipDupConfig) -> None:
+    if not config.sentry:
+        return
+    if config.sentry.debug:
+        level, event_level, attach_stacktrace = logging.DEBUG, logging.WARNING, True
+    else:
+        level, event_level, attach_stacktrace = logging.INFO, logging.ERROR, False
+
+    integrations = [
+        AioHttpIntegration(),
+        LoggingIntegration(
+            level=level,
+            event_level=event_level,
+        ),
+    ]
+    sentry_sdk.init(
+        dsn=config.sentry.dsn,
+        environment=config.sentry.environment,
+        integrations=integrations,
+        release=__version__,
+        attach_stacktrace=attach_stacktrace,
+    )
+
+
 @click.group()
 @click.version_option(__version__)
 @click.option('--config', '-c', type=str, multiple=True, help='Path to dipdup YAML config', default=['dipdup.yml'])
 @click.option('--env-file', '-e', type=str, multiple=True, help='Path to .env file', default=[])
 @click.option('--logging-config', '-l', type=str, help='Path to logging YAML config', default='logging.yml')
 @click.pass_context
+@cli_wrapper
 async def cli(ctx, config: List[str], env_file: List[str], logging_config: str):
     try:
         path = join(os.getcwd(), logging_config)
@@ -44,6 +86,7 @@ async def cli(ctx, config: List[str], env_file: List[str], logging_config: str):
         _logging_config = LoggingConfig.load(path)
     _logging_config.apply()
 
+    # NOTE: Apply env files before loading config
     for env_path in env_file:
         env_path = join(os.getcwd(), env_path)
         if not exists(env_path):
@@ -52,21 +95,14 @@ async def cli(ctx, config: List[str], env_file: List[str], logging_config: str):
         load_dotenv(env_path, override=True)
 
     _config = DipDupConfig.load(config)
+    init_sentry(_config)
+
     if _config.spec_version not in spec_version_mapping:
         raise ConfigurationError('Unknown `spec_version`, correct ones: {}')
     if _config.spec_version != __spec_version__ and ctx.invoked_subcommand != 'migrate':
         reindex = spec_reindex_mapping[__spec_version__]
         raise MigrationRequiredError(None, _config.spec_version, __spec_version__, reindex)
 
-    if _config.sentry:
-        sentry_sdk.init(
-            dsn=_config.sentry.dsn,
-            environment=_config.sentry.environment,
-            integrations=[AioHttpIntegration()],
-        )
-
-    set_decimal_context(_config.package)
-
     ctx.obj = CLIContext(
         config_paths=config,
         config=_config,
@@ -78,15 +114,18 @@ async def cli(ctx, config: List[str], env_file: List[str], logging_config: str):
 @click.option('--reindex', is_flag=True, help='Drop database and start indexing from scratch')
 @click.option('--oneshot', is_flag=True, help='Synchronize indexes wia REST and exit without starting WS connection')
 @click.pass_context
+@cli_wrapper
 async def run(ctx, reindex: bool, oneshot: bool) -> None:
     config: DipDupConfig = ctx.obj.config
     config.initialize()
+    set_decimal_context(config.package)
     dipdup = DipDup(config)
     await dipdup.run(reindex, oneshot)
 
 
 @cli.command(help='Initialize new dipdup project')
 @click.pass_context
+@cli_wrapper
 async def init(ctx):
     config: DipDupConfig = ctx.obj.config
     config.pre_initialize()
@@ -96,6 +135,7 @@ async def init(ctx):
 
 @cli.command(help='Migrate project to the new spec version')
 @click.pass_context
+@cli_wrapper
 async def migrate(ctx):
     def _bump_spec_version(spec_version: str):
         for config_path in ctx.obj.config_paths:
@@ -120,30 +160,35 @@ async def migrate(ctx):
         raise ConfigurationError('Unknown `spec_version`')
 
 
+# TODO: "cache clear"?
 @cli.command(help='Clear development request cache')
 @click.pass_context
+@cli_wrapper
 async def clear_cache(ctx):
     FileCache('dipdup', flag='cs').clear()
 
 
-@cli.group()
+@cli.group(help='Docker integration related commands')
 @click.pass_context
+@cli_wrapper
 async def docker(ctx):
     ...
 
 
-@docker.command(name='init')
+@docker.command(name='init', help='Generate Docker inventory in project directory')
 @click.option('--image', '-i', type=str, help='DipDup Docker image', default=DEFAULT_DOCKER_IMAGE)
 @click.option('--tag', '-t', type=str, help='DipDup Docker tag', default=DEFAULT_DOCKER_TAG)
 @click.option('--env-file', '-e', type=str, help='Path to env_file', default=DEFAULT_DOCKER_ENV_FILE)
 @click.pass_context
+@cli_wrapper
 async def docker_init(ctx, image: str, tag: str, env_file: str):
     config: DipDupConfig = ctx.obj.config
     await DipDupCodeGenerator(config, {}).generate_docker(image, tag, env_file)
 
 
-@cli.group()
+@cli.group(help='Hasura integration related commands')
 @click.pass_context
+@cli_wrapper
 async def hasura(ctx):
     ...
 
@@ -151,15 +196,21 @@ async def hasura(ctx):
 @hasura.command(name='configure', help='Configure Hasura GraphQL Engine')
 @click.option('--reset', is_flag=True, help='Reset metadata before configuring')
 @click.pass_context
+@cli_wrapper
 async def hasura_configure(ctx, reset: bool):
     config: DipDupConfig = ctx.obj.config
     url = config.database.connection_string
     models = f'{config.package}.models'
     if not config.hasura:
-        _logger.error('`hasura` config section is empty')
-        return
-    hasura_gateway = HasuraGateway(config.package, config.hasura, cast(PostgresDatabaseConfig, config.database))
+        raise ConfigurationError('`hasura` config section is empty')
+    if reset:
+        config.hasura.reset = True
+    hasura_gateway = HasuraGateway(
+        package=config.package,
+        hasura_config=config.hasura,
+        database_config=cast(PostgresDatabaseConfig, config.database),
+    )
 
     async with tortoise_wrapper(url, models):
         async with hasura_gateway:
-            await hasura_gateway.configure(reset)
+            await hasura_gateway.configure()
