@@ -7,10 +7,11 @@ import re
 import sys
 from abc import ABC, abstractmethod
 from collections import defaultdict
+from copy import copy
 from enum import Enum
 from os import environ as env
 from os.path import dirname
-from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union, cast
+from typing import Any, Callable, Dict, List, Optional, Sequence, Set, Type, Union, cast
 from urllib.parse import urlparse
 
 from pydantic import Field, validator
@@ -19,8 +20,8 @@ from pydantic.json import pydantic_encoder
 from ruamel.yaml import YAML
 from typing_extensions import Literal
 
-from dipdup.exceptions import ConfigurationError, HandlerImportError
-from dipdup.utils import pascal_to_snake, snake_to_pascal
+from dipdup.exceptions import ConfigurationError
+from dipdup.utils import import_from, pascal_to_snake, snake_to_pascal
 
 ROLLBACK_HANDLER = 'on_rollback'
 CONFIGURE_HANDLER = 'on_configure'
@@ -36,6 +37,7 @@ _logger = logging.getLogger('dipdup.config')
 class OperationType(Enum):
     transaction = 'transaction'
     origination = 'origination'
+    migration = 'migration'
 
 
 @dataclass
@@ -100,19 +102,35 @@ class HTTPConfig:
     connection_limit: Optional[int] = None
     batch_size: Optional[int] = None
 
-    def merge(self, other: Optional['HTTPConfig']) -> None:
-        if not other:
-            return
-        for k, v in other.__dict__.items():
-            if v is not None:
-                setattr(self, k, v)
+    def merge(self, other: Optional['HTTPConfig']) -> 'HTTPConfig':
+        config = copy(self)
+        if other:
+            for k, v in other.__dict__.items():
+                if v is not None:
+                    setattr(config, k, v)
+        return config
 
 
 @dataclass
-class ContractConfig:
+class NameMixin:
+    def __post_init_post_parse__(self) -> None:
+        self._name: Optional[str] = None
+
+    @property
+    def name(self) -> str:
+        if self._name is None:
+            raise RuntimeError('Config is not pre-initialized')
+        return self._name
+
+    @name.setter
+    def name(self, name: str) -> None:
+        self._name = name
+
+
+@dataclass
+class ContractConfig(NameMixin):
     """Contract config
 
-    :param network: Corresponding network alias, only for sanity checks
     :param address: Contract address
     :param typename: User-defined alias for the contract script
     """
@@ -135,22 +153,6 @@ class ContractConfig:
         return v
 
 
-@dataclass
-class NameMixin:
-    def __post_init_post_parse__(self) -> None:
-        self._name: Optional[str] = None
-
-    @property
-    def name(self) -> str:
-        if self._name is None:
-            raise RuntimeError('Config is not pre-initialized')
-        return self._name
-
-    @name.setter
-    def name(self, name: str) -> None:
-        self._name = name
-
-
 @dataclass
 class TzktDatasourceConfig(NameMixin):
     """TzKT datasource config
@@ -277,12 +279,27 @@ class StorageTypeMixin:
 
     def initialize_storage_cls(self, package: str, module_name: str) -> None:
         _logger.info('Registering `%s` storage type', module_name)
-        storage_type_module = importlib.import_module(f'{package}.types.{module_name}.storage')
-        storage_type_cls = getattr(
-            storage_type_module,
-            snake_to_pascal(module_name) + 'Storage',
-        )
-        self.storage_type_cls = storage_type_cls
+        cls_name = snake_to_pascal(module_name) + 'Storage'
+        module_name = f'{package}.types.{module_name}.storage'
+        self.storage_type_cls = import_from(module_name, cls_name)
+
+
+@dataclass
+class ParentMixin:
+    """`parent` field for index and template configs"""
+
+    def __post_init_post_parse__(self):
+        self._parent: Optional['IndexConfig'] = None
+
+    @property
+    def parent(self) -> Optional['IndexConfig']:
+        return self._parent
+
+    @parent.setter
+    def parent(self, config: 'IndexConfig') -> None:
+        if self._parent:
+            raise RuntimeError('Can\'t unset parent once set')
+        self._parent = config
 
 
 @dataclass
@@ -304,9 +321,9 @@ class ParameterTypeMixin:
 
     def initialize_parameter_cls(self, package: str, module_name: str, entrypoint: str) -> None:
         _logger.info('Registering parameter type for entrypoint `%s`', entrypoint)
-        parameter_type_module = importlib.import_module(f'{package}.types.{module_name}.parameter.{pascal_to_snake(entrypoint)}')
-        parameter_type_cls = getattr(parameter_type_module, snake_to_pascal(entrypoint) + 'Parameter')
-        self.parameter_type_cls = parameter_type_cls
+        module_name = f'{package}.types.{module_name}.parameter.{pascal_to_snake(entrypoint)}'
+        cls_name = snake_to_pascal(entrypoint) + 'Parameter'
+        self.parameter_type_cls = import_from(module_name, cls_name)
 
 
 @dataclass
@@ -459,10 +476,11 @@ class OperationHandlerOriginationPatternConfig(PatternConfig, StorageTypeMixin):
 
 
 @dataclass
-class HandlerConfig:
+class HandlerConfig(NameMixin):
     callback: str
 
     def __post_init_post_parse__(self):
+        super().__post_init_post_parse__()
         self._callback_fn = None
         if self.callback in (ROLLBACK_HANDLER, CONFIGURE_HANDLER):
             raise ConfigurationError(f'`{self.callback}` callback name is reserved')
@@ -497,10 +515,10 @@ class OperationHandlerConfig(HandlerConfig):
 @dataclass
 class TemplateValuesMixin:
     def __post_init_post_parse__(self) -> None:
-        self._template_values: Optional[Dict[str, str]] = None
+        self._template_values: Dict[str, str] = {}
 
     @property
-    def template_values(self) -> Optional[Dict[str, str]]:
+    def template_values(self) -> Dict[str, str]:
         return self._template_values
 
     @template_values.setter
@@ -509,12 +527,13 @@ class TemplateValuesMixin:
 
 
 @dataclass
-class IndexConfig(TemplateValuesMixin, NameMixin):
+class IndexConfig(TemplateValuesMixin, NameMixin, ParentMixin):
     datasource: Union[str, TzktDatasourceConfig]
 
     def __post_init_post_parse__(self) -> None:
         TemplateValuesMixin.__post_init_post_parse__(self)
         NameMixin.__post_init_post_parse__(self)
+        ParentMixin.__post_init_post_parse__(self)
 
     def hash(self) -> str:
         config_json = json.dumps(self, default=pydantic_encoder)
@@ -532,10 +551,11 @@ class IndexConfig(TemplateValuesMixin, NameMixin):
 class OperationIndexConfig(IndexConfig):
     """Operation index config
 
-    :param datasource: Alias of datasource in `datasources` block
-    :param contract: Alias of contract to fetch operations for
-    :param first_block: First block to process
-    :param last_block: Last block to process
+    :param datasource: Alias of index datasource in `datasources` section
+    :param contracts: Aliases of contracts being indexed in `contracts` section
+    :param stateless: Makes index dynamic. DipDup will synchronize index from the first block on every run
+    :param first_block: First block to process (use with `--oneshot` run argument)
+    :param last_block: Last block to process (use with `--oneshot` run argument)
     :param handlers: List of indexer handlers
     """
 
@@ -557,6 +577,15 @@ class OperationIndexConfig(IndexConfig):
                 raise RuntimeError('Config is not initialized')
         return cast(List[ContractConfig], self.contracts)
 
+    @property
+    def entrypoints(self) -> Set[str]:
+        entrypoints = set()
+        for handler in self.handlers:
+            for pattern in handler.pattern:
+                if isinstance(pattern, OperationHandlerTransactionPatternConfig) and pattern.entrypoint:
+                    entrypoints.add(pattern.entrypoint)
+        return entrypoints
+
 
 @dataclass
 class BigMapHandlerConfig(HandlerConfig):
@@ -605,34 +634,20 @@ class BigMapIndexConfig(IndexConfig):
     first_block: int = 0
     last_block: int = 0
 
-
-@dataclass
-class BlockHandlerConfig(HandlerConfig):
-    pattern = None
+    @property
+    def contracts(self) -> List[ContractConfig]:
+        return list(set([cast(ContractConfig, handler_config.contract) for handler_config in self.handlers]))
 
 
 @dataclass
-class BlockIndexConfig(IndexConfig):
-    """Stub, not implemented"""
-
-    kind: Literal['block']
-    datasource: Union[str, TzktDatasourceConfig]
-    handlers: List[BlockHandlerConfig]
-
-    stateless: bool = False
-    first_block: int = 0
-    last_block: int = 0
-
-
-@dataclass
-class StaticTemplateConfig:
+class IndexTemplateConfig(ParentMixin):
     kind = 'template'
     template: str
     values: Dict[str, str]
 
 
-IndexConfigT = Union[OperationIndexConfig, BigMapIndexConfig, BlockIndexConfig, StaticTemplateConfig]
-IndexConfigTemplateT = Union[OperationIndexConfig, BigMapIndexConfig, BlockIndexConfig]
+IndexConfigT = Union[OperationIndexConfig, BigMapIndexConfig, IndexTemplateConfig]
+IndexConfigTemplateT = Union[OperationIndexConfig, BigMapIndexConfig]
 HandlerPatternConfigT = Union[OperationHandlerOriginationPatternConfig, OperationHandlerTransactionPatternConfig]
 
 
@@ -646,6 +661,7 @@ class HasuraConfig:
     camel_case: bool = False
     connection_timeout: int = 5
     rest: bool = True
+    reset: Optional[bool] = None
     http: Optional[HTTPConfig] = None
 
     @validator('url', allow_reuse=True)
@@ -684,6 +700,7 @@ class JobConfig(HandlerConfig):
 class SentryConfig:
     dsn: str
     environment: Optional[str] = None
+    debug: bool = False
 
 
 @dataclass
@@ -705,22 +722,14 @@ class DipDupConfig:
     spec_version: str
     package: str
     datasources: Dict[str, DatasourceConfigT]
+    database: Union[SqliteDatabaseConfig, PostgresDatabaseConfig] = SqliteDatabaseConfig(kind='sqlite')
     contracts: Dict[str, ContractConfig] = Field(default_factory=dict)
     indexes: Dict[str, IndexConfigT] = Field(default_factory=dict)
-    templates: Optional[Dict[str, IndexConfigTemplateT]] = None
-    database: Union[SqliteDatabaseConfig, PostgresDatabaseConfig] = SqliteDatabaseConfig(kind='sqlite')
+    templates: Dict[str, IndexConfigTemplateT] = Field(default_factory=dict)
+    jobs: Dict[str, JobConfig] = Field(default_factory=dict)
     hasura: Optional[HasuraConfig] = None
-    jobs: Optional[Dict[str, JobConfig]] = None
     sentry: Optional[SentryConfig] = None
 
-    @property
-    def environment(self) -> Dict[str, str]:
-        return self._environment
-
-    @property
-    def filenames(self) -> List[str]:
-        return self._filenames
-
     def __post_init_post_parse__(self):
         self._filenames: List[str] = []
         self._environment: Dict[str, str] = {}
@@ -729,25 +738,34 @@ class DipDupConfig:
         self._initialized = []
         self.validate()
 
+    @property
+    def environment(self) -> Dict[str, str]:
+        return self._environment
+
+    @property
+    def filenames(self) -> List[str]:
+        return self._filenames
+
     def validate(self) -> None:
         if isinstance(self.database, SqliteDatabaseConfig) and self.hasura:
             raise ConfigurationError('SQLite DB engine is not supported by Hasura')
 
     def get_contract(self, name: str) -> ContractConfig:
+        self._check_name(name)
         try:
             return self.contracts[name]
         except KeyError as e:
             raise ConfigurationError(f'Contract `{name}` not found in `contracts` config section') from e
 
     def get_datasource(self, name: str) -> DatasourceConfigT:
+        self._check_name(name)
         try:
             return self.datasources[name]
         except KeyError as e:
             raise ConfigurationError(f'Datasource `{name}` not found in `datasources` config section') from e
 
     def get_template(self, name: str) -> IndexConfigTemplateT:
-        if not self.templates:
-            raise ConfigurationError('`templates` section is missing')
+        self._check_name(name)
         try:
             return self.templates[name]
         except KeyError as e:
@@ -760,23 +778,19 @@ class DipDupConfig:
         return datasource
 
     def get_rollback_fn(self) -> Type:
-        try:
-            module_name = f'{self.package}.handlers.{ROLLBACK_HANDLER}'
-            return getattr(importlib.import_module(module_name), ROLLBACK_HANDLER)
-        except (ModuleNotFoundError, AttributeError) as e:
-            raise HandlerImportError(module=module_name, obj=ROLLBACK_HANDLER) from e
+        module_name = f'{self.package}.handlers.{ROLLBACK_HANDLER}'
+        fn_name = ROLLBACK_HANDLER
+        return import_from(module_name, fn_name)
 
     def get_configure_fn(self) -> Type:
-        try:
-            module_name = f'{self.package}.handlers.{CONFIGURE_HANDLER}'
-            return getattr(importlib.import_module(module_name), CONFIGURE_HANDLER)
-        except (ModuleNotFoundError, AttributeError) as e:
-            raise HandlerImportError(module=module_name, obj=CONFIGURE_HANDLER) from e
+        module_name = f'{self.package}.handlers.{CONFIGURE_HANDLER}'
+        fn_name = CONFIGURE_HANDLER
+        return import_from(module_name, fn_name)
 
-    def resolve_static_templates(self) -> None:
+    def resolve_index_templates(self) -> None:
         _logger.info('Substituting index templates')
         for index_name, index_config in self.indexes.items():
-            if isinstance(index_config, StaticTemplateConfig):
+            if isinstance(index_config, IndexTemplateConfig):
                 template = self.get_template(index_config.template)
                 raw_template = json.dumps(template, default=pydantic_encoder)
                 for key, value in index_config.values.items():
@@ -785,8 +799,14 @@ class DipDupConfig:
                 json_template = json.loads(raw_template)
                 new_index_config = template.__class__(**json_template)
                 new_index_config.template_values = index_config.values
+                new_index_config.parent = index_config.parent
                 self.indexes[index_name] = new_index_config
 
+    def _check_name(self, name: str) -> None:
+        variable = name.split('<')[-1].split('>')[0]
+        if variable != name:
+            raise ConfigurationError(f'`{variable}` variable of index template is not set')
+
     def _pre_initialize_index(self, index_name: str, index_config: IndexConfigT) -> None:
         """Resolve contract and datasource configs by aliases"""
         if index_name in self._pre_initialized:
@@ -840,10 +860,14 @@ class DipDupConfig:
         self._pre_initialized.append(index_name)
 
     def pre_initialize(self) -> None:
-        for name, config in self.datasources.items():
-            config.name = name
+        for name, contract_config in self.contracts.items():
+            contract_config.name = name
+        for name, datasource_config in self.datasources.items():
+            datasource_config.name = name
+        for name, job_config in self.jobs.items():
+            job_config.name = name
 
-        self.resolve_static_templates()
+        self.resolve_index_templates()
         for index_name, index_config in self.indexes.items():
             self._pre_initialize_index(index_name, index_config)
 
@@ -919,29 +943,21 @@ class DipDupConfig:
 
     def _initialize_handler_callback(self, handler_config: HandlerConfig) -> None:
         _logger.info('Registering handler callback `%s`', handler_config.callback)
-        try:
-            module_name = f'{self.package}.handlers.{handler_config.callback}'
-            module = importlib.import_module(module_name)
-            callback_fn = getattr(module, handler_config.callback)
-            handler_config.callback_fn = callback_fn
-        except (ModuleNotFoundError, AttributeError) as e:
-            raise HandlerImportError(module=module_name, obj=handler_config.callback) from e
+        module_name = f'{self.package}.handlers.{handler_config.callback}'
+        fn_name = handler_config.callback
+        handler_config.callback_fn = import_from(module_name, fn_name)
 
     def _initialize_job_callback(self, job_config: JobConfig) -> None:
         _logger.info('Registering job callback `%s`', job_config.callback)
-        try:
-            module_name = f'{self.package}.jobs.{job_config.callback}'
-            module = importlib.import_module(module_name)
-            callback_fn = getattr(module, job_config.callback)
-            job_config.callback_fn = callback_fn
-        except (ModuleNotFoundError, AttributeError) as e:
-            raise HandlerImportError(module=module_name, obj=job_config.callback) from e
+        module_name = f'{self.package}.jobs.{job_config.callback}'
+        fn_name = job_config.callback
+        job_config.callback_fn = import_from(module_name, fn_name)
 
     def _initialize_index(self, index_name: str, index_config: IndexConfigT) -> None:
         if index_name in self._initialized:
             return
 
-        if isinstance(index_config, StaticTemplateConfig):
+        if isinstance(index_config, IndexTemplateConfig):
             raise RuntimeError('Config is not pre-initialized')
 
         if isinstance(index_config, OperationIndexConfig):
