@@ -1,5 +1,7 @@
+from abc import abstractmethod
 from enum import Enum
-from typing import Awaitable, List, Optional, Protocol
+from logging import Logger
+from typing import Awaitable, List, Protocol
 
 from pyee import AsyncIOEventEmitter  # type: ignore
 
@@ -21,7 +23,7 @@ class OperationsCallback(Protocol):
 
 
 class BigMapsCallback(Protocol):
-    def __call__(self, datasource: 'IndexDatasource', big_maps: List[BigMapData]) -> Awaitable[None]:
+    def __call__(self, datasource: 'IndexDatasource', big_maps: List[BigMapData], block: HeadBlockData) -> Awaitable[None]:
         ...
 
 
@@ -35,8 +37,16 @@ class HeadCallback(Protocol):
         ...
 
 
-class IndexDatasource(HTTPGateway, AsyncIOEventEmitter):
-    def __init__(self, url: str, http_config: Optional[HTTPConfig] = None) -> None:
+class Datasource(HTTPGateway):
+    _logger: Logger
+
+    @abstractmethod
+    async def run(self) -> None:
+        ...
+
+
+class IndexDatasource(Datasource, AsyncIOEventEmitter):
+    def __init__(self, url: str, http_config: HTTPConfig) -> None:
         HTTPGateway.__init__(self, url, http_config)
         AsyncIOEventEmitter.__init__(self)
 
