@@ -1,10 +1,12 @@
 import asyncio
 import logging
+from collections import defaultdict
 from datetime import datetime, timezone
 from decimal import Decimal
 from enum import Enum
-from typing import Any, AsyncGenerator, Dict, List, NoReturn, Optional, Set, Tuple, cast
+from typing import Any, AsyncGenerator, DefaultDict, Dict, List, NoReturn, Optional, Set, Tuple, cast
 
+from aiohttp import ClientResponseError
 from aiosignalrcore.hub.base_hub_connection import BaseHubConnection  # type: ignore
 from aiosignalrcore.hub_connection_builder import HubConnectionBuilder  # type: ignore
 from aiosignalrcore.messages.completion_message import CompletionMessage  # type: ignore
@@ -21,7 +23,7 @@ from dipdup.config import (
 from dipdup.datasources.datasource import IndexDatasource
 from dipdup.datasources.tzkt.enums import TzktMessageType
 from dipdup.models import BigMapAction, BigMapData, BlockData, HeadBlockData, OperationData
-from dipdup.utils import split_by_chunks
+from dipdup.utils import groupby, split_by_chunks
 
 OperationID = int
 
@@ -43,6 +45,15 @@ OPERATION_FIELDS = (
     "hasInternals",
     "diffs",
 )
+ORIGINATION_MIGRATION_FIELDS = (
+    "id",
+    "level",
+    "timestamp",
+    "storage",
+    "diffs",
+    "account",
+    "balanceChange",
+)
 ORIGINATION_OPERATION_FIELDS = (
     *OPERATION_FIELDS,
     "originatedContract",
@@ -83,6 +94,7 @@ class OperationFetcher:
         transaction_addresses: Set[str],
         origination_addresses: Set[str],
         cache: bool = False,
+        migration_originations: List[OperationData] = None,
     ) -> None:
         self._datasource = datasource
         self._first_level = first_level
@@ -96,7 +108,12 @@ class OperationFetcher:
         self._heads: Dict[OperationFetcherChannel, int] = {}
         self._offsets: Dict[OperationFetcherChannel, int] = {}
         self._fetched: Dict[OperationFetcherChannel, bool] = {}
-        self._operations: Dict[int, List[OperationData]] = {}
+
+        self._operations: DefaultDict[int, List[OperationData]]
+        if migration_originations:
+            self._operations = groupby(migration_originations, lambda op: op.level)
+        else:
+            self._operations = defaultdict(list)
 
     def _get_operations_head(self, operations: List[OperationData]) -> int:
         """Get latest block level (head) of sorted operations batch"""
@@ -271,17 +288,26 @@ class TzktDatasource(IndexDatasource):
     * Calls Matchers to match received operation groups with indexes' pattern and spawn callbacks on match
     """
 
+    _default_http_config = HTTPConfig(
+        cache=True,
+        retry_sleep=1,
+        retry_multiplier=1.1,
+        ratelimit_rate=100,
+        ratelimit_period=30,
+        connection_limit=25,
+    )
+
     def __init__(
         self,
         url: str,
         http_config: Optional[HTTPConfig] = None,
     ) -> None:
-        super().__init__(url, http_config)
+        super().__init__(url, self._default_http_config.merge(http_config))
         self._logger = logging.getLogger('dipdup.tzkt')
+
         self._transaction_subscriptions: Set[str] = set()
         self._origination_subscriptions: bool = False
-        self._big_map_subscriptions: Dict[str, List[str]] = {}
-
+        self._big_map_subscriptions: Dict[str, Set[str]] = {}
         self._client: Optional[BaseHubConnection] = None
 
         self._block: Optional[HeadBlockData] = None
@@ -303,7 +329,7 @@ class TzktDatasource(IndexDatasource):
     @property
     def block(self) -> HeadBlockData:
         if self._block is None:
-            raise RuntimeError('No message from `head` channel received')
+            raise RuntimeError('Attempt to access head block before the first message')
         return self._block
 
     async def get_similar_contracts(self, address: str, strict: bool = False) -> List[str]:
@@ -378,6 +404,33 @@ class TzktDatasource(IndexDatasource):
         )
         return self.convert_block(block_json)
 
+    async def get_migration_originations(self, first_level: int = 0) -> List[OperationData]:
+        """Get contracts originated from migrations"""
+        self._logger.info('Fetching contracts originated with migrations')
+        # NOTE: Empty unwrapped request to ensure API supports migration originations
+        try:
+            await self._http._request(
+                'get',
+                url='v1/operations/migrations',
+                params={
+                    'kind': 'origination',
+                    'limit': 0,
+                },
+            )
+        except ClientResponseError:
+            return []
+
+        raw_migrations = await self._http.request(
+            'get',
+            url='v1/operations/migrations',
+            params={
+                'kind': 'origination',
+                'level.gt': first_level,
+                'select': ','.join(ORIGINATION_MIGRATION_FIELDS),
+            },
+        )
+        return [self.convert_migration_origination(m) for m in raw_migrations]
+
     async def get_originations(
         self, addresses: Set[str], offset: int, first_level: int, last_level: int, cache: bool = False
     ) -> List[OperationData]:
@@ -454,27 +507,23 @@ class TzktDatasource(IndexDatasource):
         return big_maps
 
     async def add_index(self, index_config: IndexConfigTemplateT) -> None:
-        """Register index config in internal mappings and matchers. Find and register subscriptions.
-        If called in runtime need to `resync` then."""
+        """Register index config in internal mappings and matchers. Find and register subscriptions."""
 
         if isinstance(index_config, OperationIndexConfig):
-
             for contract_config in index_config.contracts or []:
                 self._transaction_subscriptions.add(cast(ContractConfig, contract_config).address)
-
             for handler_config in index_config.handlers:
                 for pattern_config in handler_config.pattern:
                     if isinstance(pattern_config, OperationHandlerOriginationPatternConfig):
                         self._origination_subscriptions = True
 
         elif isinstance(index_config, BigMapIndexConfig):
-
             for big_map_handler_config in index_config.handlers:
                 address, path = big_map_handler_config.contract_config.address, big_map_handler_config.path
                 if address not in self._big_map_subscriptions:
-                    self._big_map_subscriptions[address] = []
+                    self._big_map_subscriptions[address] = set()
                 if path not in self._big_map_subscriptions[address]:
-                    self._big_map_subscriptions[address].append(path)
+                    self._big_map_subscriptions[address].add(path)
 
         else:
             raise NotImplementedError(f'Index kind `{index_config.kind}` is not supported')
@@ -508,6 +557,7 @@ class TzktDatasource(IndexDatasource):
 
     async def run(self) -> None:
         """Main loop. Sync indexes via REST, start WS connection"""
+        # TODO: Update docs, honor `realtime` flag
         self._logger.info('Starting datasource')
 
         self._logger.info('Starting websocket client')
@@ -519,20 +569,23 @@ class TzktDatasource(IndexDatasource):
             return
 
         self._logger.info('Connected to server')
-        await self.subscribe_to_head()
+        await self._subscribe_to_head()
         for address in self._transaction_subscriptions:
-            await self.subscribe_to_transactions(address)
+            await self._subscribe_to_transactions(address)
         # NOTE: All originations are passed to matcher
         if self._origination_subscriptions:
-            await self.subscribe_to_originations()
+            await self._subscribe_to_originations()
         for address, paths in self._big_map_subscriptions.items():
-            await self.subscribe_to_big_maps(address, paths)
+            await self._subscribe_to_big_maps(address, paths)
 
+    # NOTE: Pay attention: this is not a pyee callback
     def _on_error(self, message: CompletionMessage) -> NoReturn:
         """Raise exception from WS server's error message"""
         raise Exception(message.error)
 
-    async def subscribe_to_transactions(self, address: str) -> None:
+    # TODO: Catch exceptions from pyee 'error' channel
+
+    async def _subscribe_to_transactions(self, address: str) -> None:
         """Subscribe to contract's operations on established WS connection"""
         self._logger.info('Subscribing to %s transactions', address)
         await self._send(
@@ -545,7 +598,7 @@ class TzktDatasource(IndexDatasource):
             ],
         )
 
-    async def subscribe_to_originations(self) -> None:
+    async def _subscribe_to_originations(self) -> None:
         """Subscribe to all originations on established WS connection"""
         self._logger.info('Subscribing to originations')
         await self._send(
@@ -557,7 +610,7 @@ class TzktDatasource(IndexDatasource):
             ],
         )
 
-    async def subscribe_to_big_maps(self, address: str, paths: List[str]) -> None:
+    async def _subscribe_to_big_maps(self, address: str, paths: Set[str]) -> None:
         """Subscribe to contract's big map diffs on established WS connection"""
         self._logger.info('Subscribing to big map updates of %s, %s', address, paths)
         for path in paths:
@@ -571,7 +624,7 @@ class TzktDatasource(IndexDatasource):
                 ],
             )
 
-    async def subscribe_to_head(self) -> None:
+    async def _subscribe_to_head(self) -> None:
         """Subscribe to head on established WS connection"""
         self._logger.info('Subscribing to head')
         await self._send(
@@ -579,97 +632,56 @@ class TzktDatasource(IndexDatasource):
             [],
         )
 
-    def _default_http_config(self) -> HTTPConfig:
-        return HTTPConfig(
-            cache=True,
-            retry_sleep=1,
-            retry_multiplier=1.1,
-            ratelimit_rate=100,
-            ratelimit_period=30,
-            connection_limit=25,
-        )
+    async def _extract_message_data(self, channel: str, message: List[Any]) -> Any:
+        for item in message:
+            head_level = item['state']
+            message_type = TzktMessageType(item['type'])
+            self._logger.debug('`%s` message: %s', channel, message_type.name)
+
+            if message_type == TzktMessageType.STATE:
+                if self._sync_level != head_level:
+                    self._logger.info('Datasource level set to %s', head_level)
+                    self._sync_level = head_level
+                    self._level = head_level
+
+            elif message_type == TzktMessageType.DATA:
+                self._level = head_level
+                yield item['data']
+
+            elif message_type == TzktMessageType.REORG:
+                if self.level is None:
+                    raise RuntimeError
+                self.emit_rollback(self.level, head_level)
+
+            else:
+                raise NotImplementedError
 
     async def _on_operation_message(self, message: List[Dict[str, Any]]) -> None:
         """Parse and emit raw operations from WS"""
-        for item in message:
-            current_level = item['state']
-            message_type = TzktMessageType(item['type'])
-            self._logger.info('Got operation message, %s, level %s', message_type, current_level)
-
-            if message_type == TzktMessageType.STATE:
-                self._sync_level = current_level
-                self._level = current_level
-
-            elif message_type == TzktMessageType.DATA:
-                self._level = current_level
-                operations = []
-                for operation_json in item['data']:
-                    operation = self.convert_operation(operation_json)
-                    if operation.status != 'applied':
-                        continue
-                    operations.append(operation)
-                if operations:
-                    self.emit_operations(operations, self.block)
-
-            elif message_type == TzktMessageType.REORG:
-                if self.level is None:
-                    raise RuntimeError
-                self.emit_rollback(self.level, current_level)
-
-            else:
-                raise NotImplementedError
+        async for data in self._extract_message_data('operation', message):
+            operations = []
+            for operation_json in data:
+                operation = self.convert_operation(operation_json)
+                if operation.status != 'applied':
+                    continue
+                operations.append(operation)
+            if operations:
+                self.emit_operations(operations, self.block)
 
     async def _on_big_map_message(self, message: List[Dict[str, Any]]) -> None:
         """Parse and emit raw big map diffs from WS"""
-        for item in message:
-            current_level = item['state']
-            message_type = TzktMessageType(item['type'])
-            self._logger.info('Got big map message, %s, level %s', message_type, current_level)
-
-            if message_type == TzktMessageType.STATE:
-                self._sync_level = current_level
-                self._level = current_level
-
-            elif message_type == TzktMessageType.DATA:
-                self._level = current_level
-                big_maps = []
-                for big_map_json in item['data']:
-                    big_map = self.convert_big_map(big_map_json)
-                    big_maps.append(big_map)
-                self.emit_big_maps(big_maps)
-
-            elif message_type == TzktMessageType.REORG:
-                if self.level is None:
-                    raise RuntimeError
-                self.emit_rollback(self.level, current_level)
-
-            else:
-                raise NotImplementedError
+        async for data in self._extract_message_data('big_map', message):
+            big_maps = []
+            for big_map_json in data:
+                big_map = self.convert_big_map(big_map_json)
+                big_maps.append(big_map)
+            self.emit_big_maps(big_maps)
 
     async def _on_head_message(self, message: List[Dict[str, Any]]) -> None:
-        for item in message:
-            current_level = item['state']
-            message_type = TzktMessageType(item['type'])
-            self._logger.info('Got block message, %s, level %s', message_type, current_level)
-
-            if message_type == TzktMessageType.STATE:
-                self._sync_level = current_level
-                self._level = current_level
-
-            elif message_type == TzktMessageType.DATA:
-                self._level = current_level
-                block_json = item['data']
-                block = self.convert_head_block(block_json)
-                self._block = block
-                self.emit_head(block)
-
-            elif message_type == TzktMessageType.REORG:
-                if self.level is None:
-                    raise RuntimeError
-                self.emit_rollback(self.level, current_level)
-
-            else:
-                raise NotImplementedError
+        async for data in self._extract_message_data('head', message):
+            block = self.convert_head_block(data)
+            self._block = block
+            self.emit_head(block)
 
     @classmethod
     def convert_operation(cls, operation_json: Dict[str, Any]) -> OperationData:
@@ -697,8 +709,8 @@ class TzktDatasource(IndexDatasource):
             nonce=operation_json.get('nonce'),
             target_alias=operation_json['target'].get('alias') if operation_json.get('target') else None,
             initiator_alias=operation_json['initiator'].get('alias') if operation_json.get('initiator') else None,
-            entrypoint=operation_json['parameter']['entrypoint'] if operation_json.get('parameter') else None,
-            parameter_json=operation_json['parameter']['value'] if operation_json.get('parameter') else None,
+            entrypoint=operation_json['parameter'].get('entrypoint') if operation_json.get('parameter') else None,
+            parameter_json=operation_json['parameter'].get('value') if operation_json.get('parameter') else None,
             originated_contract_address=operation_json['originatedContract']['address']
             if operation_json.get('originatedContract')
             else None,
@@ -712,6 +724,35 @@ class TzktDatasource(IndexDatasource):
             diffs=operation_json.get('diffs'),
         )
 
+    @classmethod
+    def convert_migration_origination(cls, migration_origination_json: Dict[str, Any]) -> OperationData:
+        """Convert raw migration message from REST into dataclass"""
+        storage = migration_origination_json.get('storage')
+        # FIXME: Plain storage, has issues in codegen: KT1CpeSQKdkhWi4pinYcseCFKmDhs5M74BkU
+        if not isinstance(storage, Dict):
+            storage = {}
+
+        fake_operation_data = OperationData(
+            type='origination',
+            id=migration_origination_json['id'],
+            level=migration_origination_json['level'],
+            timestamp=cls._parse_timestamp(migration_origination_json['timestamp']),
+            block=migration_origination_json.get('block'),
+            originated_contract_address=migration_origination_json['account']['address'],
+            originated_contract_alias=migration_origination_json['account'].get('alias'),
+            amount=migration_origination_json['balanceChange'],
+            storage=storage,
+            diffs=migration_origination_json.get('diffs'),
+            status='applied',
+            has_internals=False,
+            hash='[none]',
+            counter=0,
+            sender_address='[none]',
+            target_address=None,
+            initiator_address=None,
+        )
+        return fake_operation_data
+
     @classmethod
     def convert_big_map(cls, big_map_json: Dict[str, Any]) -> BigMapData:
         """Convert raw big map diff message from WS/REST into dataclass"""
