@@ -1,7 +1,10 @@
 import asyncio
 import hashlib
 import logging
-from contextlib import AsyncExitStack, asynccontextmanager
+import operator
+from collections import Counter
+from contextlib import AsyncExitStack, asynccontextmanager, suppress
+from functools import reduce
 from os import listdir
 from os.path import join
 from typing import Dict, List, Optional, cast
@@ -18,25 +21,25 @@ from dipdup.config import (
     BcdDatasourceConfig,
     BigMapIndexConfig,
     CoinbaseDatasourceConfig,
+    ContractConfig,
     DatasourceConfigT,
     DipDupConfig,
     IndexConfigTemplateT,
+    IndexTemplateConfig,
     OperationIndexConfig,
     PostgresDatabaseConfig,
-    StaticTemplateConfig,
     TzktDatasourceConfig,
 )
 from dipdup.context import DipDupContext, RollbackHandlerContext
-from dipdup.datasources import DatasourceT
 from dipdup.datasources.bcd.datasource import BcdDatasource
 from dipdup.datasources.coinbase.datasource import CoinbaseDatasource
-from dipdup.datasources.datasource import IndexDatasource
+from dipdup.datasources.datasource import Datasource, IndexDatasource
 from dipdup.datasources.tzkt.datasource import TzktDatasource
 from dipdup.exceptions import ConfigurationError, ReindexingRequiredError
 from dipdup.hasura import HasuraGateway
 from dipdup.index import BigMapIndex, Index, OperationIndex
 from dipdup.models import BigMapData, HeadBlockData, IndexType, OperationData, State
-from dipdup.utils import FormattedLogger, slowdown, tortoise_wrapper
+from dipdup.utils import FormattedLogger, iter_files, slowdown, tortoise_wrapper
 
 INDEX_DISPATCHER_INTERVAL = 1.0
 from dipdup.scheduler import add_job, create_scheduler
@@ -58,7 +61,7 @@ class IndexDispatcher:
             datasource_name = cast(TzktDatasourceConfig, index_config.datasource).name
             datasource = self._ctx.datasources[datasource_name]
             if not isinstance(datasource, TzktDatasource):
-                raise RuntimeError
+                raise RuntimeError(f'`{datasource_name}` is not a TzktDatasource')
             operation_index = OperationIndex(self._ctx, index_config, datasource)
             self._indexes[index_config.name] = operation_index
             await datasource.add_index(index_config)
@@ -67,7 +70,7 @@ class IndexDispatcher:
             datasource_name = cast(TzktDatasourceConfig, index_config.datasource).name
             datasource = self._ctx.datasources[datasource_name]
             if not isinstance(datasource, TzktDatasource):
-                raise RuntimeError
+                raise RuntimeError(f'`{datasource_name}` is not a TzktDatasource')
             big_map_index = BigMapIndex(self._ctx, index_config, datasource)
             self._indexes[index_config.name] = big_map_index
             await datasource.add_index(index_config)
@@ -83,12 +86,24 @@ class IndexDispatcher:
         self._ctx.config.initialize()
 
         for index_config in self._ctx.config.indexes.values():
-            if isinstance(index_config, StaticTemplateConfig):
-                raise RuntimeError
+            if isinstance(index_config, IndexTemplateConfig):
+                raise RuntimeError('Config is not initialized')
             await self.add_index(index_config)
 
         self._ctx.reset()
 
+        contracts = [index._config.contracts for index in self._indexes.values() if index._config.contracts]
+        if not contracts:
+            return
+
+        plain_contracts = reduce(operator.add, contracts)
+        duplicate_contracts = [cast(ContractConfig, item).name for item, count in Counter(plain_contracts).items() if count > 1]
+        if duplicate_contracts:
+            self._logger.warning(
+                "The following contracts are used in more than one index: %s. Make sure you know what you're doing.",
+                ' '.join(duplicate_contracts),
+            )
+
     async def dispatch_operations(self, datasource: TzktDatasource, operations: List[OperationData], block: HeadBlockData) -> None:
         assert len(set(op.level for op in operations)) == 1
         level = operations[0].level
@@ -96,22 +111,23 @@ class IndexDispatcher:
             if isinstance(index, OperationIndex) and index.datasource == datasource:
                 index.push(level, operations, block)
 
-    async def dispatch_big_maps(self, datasource: TzktDatasource, big_maps: List[BigMapData]) -> None:
+    async def dispatch_big_maps(self, datasource: TzktDatasource, big_maps: List[BigMapData], block: HeadBlockData) -> None:
         assert len(set(op.level for op in big_maps)) == 1
         level = big_maps[0].level
         for index in self._indexes.values():
             if isinstance(index, BigMapIndex) and index.datasource == datasource:
-                index.push(level, big_maps)
+                index.push(level, big_maps, block)
 
     async def _rollback(self, datasource: TzktDatasource, from_level: int, to_level: int) -> None:
         logger = FormattedLogger(ROLLBACK_HANDLER)
         if from_level - to_level == 1:
             # NOTE: Single level rollbacks are processed at Index level.
-            # NOTE: Notify all indexes with rolled back datasource to skip next level and just verify it
+            # NOTE: Notify all indexes which use rolled back datasource to drop duplicated operations from the next block
             for index in self._indexes.values():
                 if index.datasource == datasource:
                     # NOTE: Continue to rollback with handler
                     if not isinstance(index, OperationIndex):
+                        self._logger.info('Single level rollback is not supported by `%s` indexes', index._config.kind)
                         break
                     await index.single_level_rollback(from_level)
             else:
@@ -161,8 +177,8 @@ class DipDup:
     def __init__(self, config: DipDupConfig) -> None:
         self._logger = logging.getLogger('dipdup')
         self._config = config
-        self._datasources: Dict[str, DatasourceT] = {}
-        self._datasources_by_config: Dict[DatasourceConfigT, DatasourceT] = {}
+        self._datasources: Dict[str, Datasource] = {}
+        self._datasources_by_config: Dict[DatasourceConfigT, Datasource] = {}
         self._ctx = DipDupContext(
             config=self._config,
             datasources=self._datasources,
@@ -190,7 +206,7 @@ class DipDup:
         url = self._config.database.connection_string
         models = f'{self._config.package}.models'
 
-        await self._create_datasources()
+        await self._create_datasources(realtime=not oneshot)
 
         hasura_gateway: Optional[HasuraGateway]
         if self._config.hasura:
@@ -203,15 +219,18 @@ class DipDup:
         async with AsyncExitStack() as stack:
             worker_tasks = []
             await stack.enter_async_context(tortoise_wrapper(url, models))
+
+            await self._initialize_database(reindex)
             for datasource in self._datasources.values():
                 await stack.enter_async_context(datasource)
+
+            # NOTE: on_configure hook fires after database and datasources are initialized but before Hasura is
+            await self._on_configure()
+
             if hasura_gateway:
                 await stack.enter_async_context(hasura_gateway)
                 worker_tasks.append(asyncio.create_task(hasura_gateway.configure()))
 
-            await self._initialize_database(reindex)
-            await self._configure()
-
             self._logger.info('Starting datasources')
             datasource_tasks = [] if oneshot else [asyncio.create_task(d.run()) for d in self._datasources.values()]
 
@@ -240,14 +259,14 @@ class DipDup:
         await codegen.migrate_user_handlers_to_v11()
         self._finish_migration('1.1')
 
-    async def _configure(self) -> None:
+    async def _on_configure(self) -> None:
         """Run user-defined initial configuration handler"""
         configure_fn = self._config.get_configure_fn()
         await configure_fn(self._ctx)
         self._config.initialize()
 
-    async def _create_datasources(self) -> None:
-        datasource: DatasourceT
+    async def _create_datasources(self, realtime: bool = True) -> None:
+        datasource: Datasource
         for name, datasource_config in self._config.datasources.items():
             if name in self._datasources:
                 continue
@@ -270,6 +289,7 @@ class DipDup:
             else:
                 raise NotImplementedError
 
+            datasource._logger = FormattedLogger(datasource._logger.name, datasource_config.name + ': {}')
             datasource.set_user_agent(self._config.package)
             self._datasources[name] = datasource
             self._datasources_by_config[datasource_config] = datasource
@@ -336,15 +356,11 @@ class DipDup:
         if not exists(sql_path):
             return
         self._logger.info('Executing SQL scripts from `%s`', sql_path)
-        for filename in sorted(listdir(sql_path)):
-            if not filename.endswith('.sql'):
-                continue
-
-            with open(join(sql_path, filename)) as file:
-                sql = file.read()
-
-            self._logger.info('Executing `%s`', filename)
-            await get_connection(None).execute_script(sql)
+        for file in iter_files(sql_path, '.sql'):
+            self._logger.info('Executing `%s`', file.name)
+            sql = file.read()
+            with suppress(AttributeError):
+                await get_connection(None).execute_script(sql)
 
     def _finish_migration(self, version: str) -> None:
         self._logger.warning('==================== WARNING =====================')
