@@ -1,14 +1,15 @@
 import asyncio
 import importlib
 import logging
+import re
 from contextlib import suppress
-from os import listdir
+from http import HTTPStatus
 from os.path import dirname, join
 from typing import Any, Dict, Iterator, List, Optional, Tuple
 
+import aiohttp
 import humps  # type: ignore
 from aiohttp import ClientConnectorError, ClientOSError
-from genericpath import exists
 from pydantic.dataclasses import dataclass
 from tortoise import fields
 from tortoise.transactions import get_connection
@@ -16,7 +17,32 @@ from tortoise.transactions import get_connection
 from dipdup.config import HasuraConfig, HTTPConfig, PostgresDatabaseConfig, pascal_to_snake
 from dipdup.exceptions import ConfigurationError
 from dipdup.http import HTTPGateway
-from dipdup.utils import iter_models
+from dipdup.utils import iter_files, iter_models
+
+_get_fields_query = '''
+query introspectionQuery($name: String!) {
+  __type(name: $name) {
+    kind
+    name
+    fields {
+      name
+      description
+      type {
+        name
+        kind
+        ofType {
+          name
+          kind
+        }
+      }
+    }
+  }
+}
+'''.replace(
+    '\n', ' '
+).replace(
+    '  ', ''
+)
 
 
 @dataclass
@@ -36,6 +62,12 @@ class HasuraError(RuntimeError):
 
 
 class HasuraGateway(HTTPGateway):
+    _default_http_config = HTTPConfig(
+        cache=False,
+        retry_count=3,
+        retry_sleep=1,
+    )
+
     def __init__(
         self,
         package: str,
@@ -43,19 +75,19 @@ class HasuraGateway(HTTPGateway):
         database_config: PostgresDatabaseConfig,
         http_config: Optional[HTTPConfig] = None,
     ) -> None:
-        super().__init__(hasura_config.url, http_config)
+        super().__init__(hasura_config.url, self._default_http_config.merge(http_config))
         self._logger = logging.getLogger('dipdup.hasura')
         self._package = package
         self._hasura_config = hasura_config
         self._database_config = database_config
 
-    async def configure(self, reset: bool = False) -> None:
+    async def configure(self) -> None:
         """Generate Hasura metadata and apply to instance with credentials from `hasura` config section."""
 
         self._logger.info('Configuring Hasura')
         await self._healthcheck()
 
-        if reset:
+        if self._hasura_config.reset is True:
             await self._reset_metadata()
 
         metadata = await self._fetch_metadata()
@@ -108,15 +140,6 @@ class HasuraGateway(HTTPGateway):
 
         self._logger.info('Hasura instance has been configured')
 
-    def _default_http_config(self) -> HTTPConfig:
-        return HTTPConfig(
-            cache=False,
-            retry_sleep=1,
-            retry_multiplier=1.1,
-            ratelimit_rate=100,
-            ratelimit_period=1,
-        )
-
     async def _hasura_request(self, endpoint: str, json: Dict[str, Any]) -> Dict[str, Any]:
         self._logger.debug('Sending `%s` request: %s', endpoint, json)
         result = await self._http.request(
@@ -164,20 +187,30 @@ class HasuraGateway(HTTPGateway):
 
     async def _replace_metadata(self, metadata: Dict[str, Any]) -> None:
         self._logger.info('Replacing metadata')
-        await self._hasura_request(
-            endpoint='query',
-            json={
-                "type": "replace_metadata",
-                "args": metadata,
-            },
-        )
+        endpoint, json = 'query', {
+            "type": "replace_metadata",
+            "args": metadata,
+        }
+        try:
+            await self._hasura_request(endpoint, json)
+        except aiohttp.ClientResponseError as e:
+            # NOTE: 400 from Hasura means we failed either to generate or to merge existing metadata.
+            # NOTE: Reset metadata and retry if not forbidden by config.
+            print(e.status, self._hasura_config.reset)
+            if e.status != HTTPStatus.BAD_REQUEST or self._hasura_config.reset is False:
+                print('raise')
+                raise
+            self._logger.warning('Failed to replace metadata, resetting')
+            await self._reset_metadata()
+            await self._hasura_request(endpoint, json)
 
     async def _get_views(self) -> List[str]:
         return [
             row[0]
             for row in (
                 await get_connection(None).execute_query(
-                    f"SELECT table_name FROM information_schema.views WHERE table_schema = '{self._database_config.schema_name}'"
+                    f"SELECT table_name FROM information_schema.views WHERE table_schema = '{self._database_config.schema_name}' UNION "
+                    f"SELECT matviewname as table_name FROM pg_matviews WHERE schemaname = '{self._database_config.schema_name}'"
                 )
             )[1]
         ]
@@ -230,14 +263,8 @@ class HasuraGateway(HTTPGateway):
         package = importlib.import_module(self._package)
         package_path = dirname(package.__file__)
         graphql_path = join(package_path, 'graphql')
-        if not exists(graphql_path):
-            return
-        for filename in sorted(listdir(graphql_path)):
-            if not filename.endswith('.graphql'):
-                continue
-
-            with open(join(graphql_path, filename)) as file:
-                yield filename[:-8], file.read()
+        for file in iter_files(graphql_path, '.graphql'):
+            yield file.name.split('/')[-1][:-8], file.read()
 
     async def _generate_query_collections_metadata(self) -> List[Dict[str, Any]]:
         queries = []
@@ -271,43 +298,30 @@ class HasuraGateway(HTTPGateway):
         generated_dict = {key(t): t for t in generated}
         return list({**existing_dict, **generated_dict}.values())
 
-    async def _get_fields(self, name: str = 'query_root') -> List[Field]:
-        query = '''
-query introspectionQuery($name: String!) {
-  __type(name: $name) {
-    kind
-    name
-    fields {
-      name
-      description
-      type {
-        name
-        kind
-        ofType {
-          name
-          kind
-        }
-      }
-    }
-  }
-}
-        '''.replace(
-            '\n', ' '
-        ).replace(
-            '  ', ''
-        )
+    async def _get_fields_json(self, name: str) -> List[Dict[str, Any]]:
         result = await self._hasura_request(
             endpoint='graphql',
             json={
-                'query': query,
+                'query': _get_fields_query,
                 'variables': {'name': name},
             },
         )
         try:
-            fields_json = result['data']['__type']['fields']
+            return result['data']['__type']['fields']
         except TypeError as e:
             raise HasuraError(f'Unknown table `{name}`') from e
 
+    async def _get_fields(self, name: str = 'query_root') -> List[Field]:
+
+        try:
+            fields_json = await self._get_fields_json(name)
+        except HasuraError:
+            # NOTE: An issue with decamelizing the table name?
+            # NOTE: dex_quotes_15m -> dexQuotes15m -> dex_quotes15m -> FAIL
+            # NOTE: Let's prefix every numeric with underscore. Won't help in complex cases but worth a try.
+            alternative_name = ''.join([f"_{w}" if w.isnumeric() else w for w in re.split(r'(\d+)', name)])
+            fields_json = await self._get_fields_json(alternative_name)
+
         fields = []
         for field_json in fields_json:
             # NOTE: Exclude autogenerated aggregate and pk fields
