@@ -1,4 +1,4 @@
-import logging
+import time
 from abc import abstractmethod
 from collections import defaultdict, deque, namedtuple
 from contextlib import suppress
@@ -36,21 +36,22 @@ class Index:
         self._config = config
         self._datasource = datasource
 
-        self._logger = logging.getLogger('dipdup.index')
+        self._logger = FormattedLogger('dipdup.index', fmt=f'{config.name}: ' + '{}')
         self._state: Optional[State] = None
 
     @property
     def datasource(self) -> TzktDatasource:
         return self._datasource
 
-    async def get_state(self) -> State:
-        """Get state of index containing current level and config hash"""
+    @property
+    def state(self) -> State:
         if self._state is None:
-            await self._initialize_index_state()
-        return cast(State, self._state)
+            raise RuntimeError('Index state is not initialized')
+        return self._state
 
     async def process(self) -> None:
-        state = await self.get_state()
+        await self._initialize_index_state()
+
         if self._config.last_block:
             last_level = self._config.last_block
             await self._synchronize(last_level, cache=True)
@@ -58,7 +59,7 @@ class Index:
             self._logger.info('Datasource is not active, sync to the latest block')
             last_level = (await self._datasource.get_head_block()).level
             await self._synchronize(last_level)
-        elif self._datasource.sync_level > state.level:
+        elif self._datasource.sync_level > self.state.level:
             self._logger.info('Index is behind datasource, sync to datasource level')
             self._queue.clear()
             last_level = self._datasource.sync_level
@@ -74,8 +75,26 @@ class Index:
     async def _process_queue(self) -> None:
         ...
 
+    async def _enter_sync_state(self, last_level: int) -> Optional[int]:
+        first_level = self.state.level
+        if first_level == last_level:
+            return None
+        if first_level > last_level:
+            raise RuntimeError(f'Attempt to synchronize index from level {first_level} to level {last_level}')
+        self._logger.info('Synchronizing index to level %s', last_level)
+        self.state.hash = None  # type: ignore
+        await self.state.save()
+        return first_level
+
+    async def _exit_sync_state(self, last_level: int) -> None:
+        self._logger.info('Index is synchronized to level %s', last_level)
+        self.state.level = last_level  # type: ignore
+        await self.state.save()
+
     async def _initialize_index_state(self) -> None:
-        self._logger.info('Getting state for index `%s`', self._config.name)
+        if self._state:
+            return
+        self._logger.info('Initializing index state')
         index_config_hash = self._config.hash()
         state = await State.get_or_none(
             index_name=self._config.name,
@@ -94,15 +113,13 @@ class Index:
             self._logger.warning('Config hash mismatch (config has been changed), reindexing')
             await self._ctx.reindex()
 
-        # NOTE: No need to check genesis block
-        if state.level:
+        self._logger.info('%s', f'{state.level=} {state.hash=}'.replace('state.', ''))
+        # NOTE: No need to check indexes which are not synchronized.
+        if state.level and state.hash:
             block = await self._datasource.get_block(state.level)
-            if state.hash:
-                if state.hash != block.hash:
-                    self._logger.warning('Block hash mismatch (missed rollback while dipdup was stopped), reindexing')
-                    await self._ctx.reindex()
-            else:
-                state.hash = block.hash  # type: ignore
+            if state.hash != block.hash:
+                self._logger.warning('Block hash mismatch (missed rollback while dipdup was stopped), reindexing')
+                await self._ctx.reindex()
 
         await state.save()
         self._state = state
@@ -114,16 +131,30 @@ class OperationIndex(Index):
     def __init__(self, ctx: DipDupContext, config: OperationIndexConfig, datasource: TzktDatasource) -> None:
         super().__init__(ctx, config, datasource)
         self._queue: Deque[Tuple[int, List[OperationData], Optional[HeadBlockData]]] = deque()
-        self._contract_hashes: Dict[str, Tuple[str, str]] = {}
+        self._contract_hashes: Dict[str, Tuple[int, int]] = {}
         self._rollback_level: Optional[int] = None
         self._last_hashes: Set[str] = set()
+        self._migration_originations: Optional[Dict[str, OperationData]] = None
 
     def push(self, level: int, operations: List[OperationData], block: Optional[HeadBlockData] = None) -> None:
         self._queue.append((level, operations, block))
 
     async def single_level_rollback(self, from_level: int) -> None:
-        """Ensure next arrived block is the same as rolled back one"""
-        self._rollback_level = from_level
+        """Ensure next arrived block is the same as rolled back one
+
+        Called by IndexDispatcher in case index datasource reported a rollback.
+        """
+        if self._rollback_level:
+            raise RuntimeError('Already in rollback state')
+
+        await self._initialize_index_state()
+        if self.state.level < from_level:
+            self._logger.info('Index level is lower than rollback level, ignoring')
+        elif self.state.level == from_level:
+            self._logger.info('Single level rollback has been triggered')
+            self._rollback_level = from_level
+        else:
+            raise RuntimeError('Index level is higher than rollback level')
 
     async def _process_queue(self) -> None:
         if not self._queue:
@@ -136,18 +167,21 @@ class OperationIndex(Index):
 
     async def _synchronize(self, last_level: int, cache: bool = False) -> None:
         """Fetch operations via Fetcher and pass to message callback"""
-        state = await self.get_state()
-        first_level = state.level
-        if first_level == last_level:
+        first_level = await self._enter_sync_state(last_level)
+        if first_level is None:
             return
-        if first_level > last_level:
-            raise RuntimeError(first_level, last_level)
 
         self._logger.info('Fetching operations from level %s to %s', first_level, last_level)
-
         transaction_addresses = await self._get_transaction_addresses()
         origination_addresses = await self._get_origination_addresses()
 
+        migration_originations = []
+        if self._config.types and OperationType.migration in self._config.types:
+            migration_originations = await self._datasource.get_migration_originations(first_level)
+            for op in migration_originations:
+                code_hash, type_hash = await self._get_contract_hashes(cast(str, op.originated_contract_address))
+                op.originated_contract_code_hash, op.originated_contract_type_hash = code_hash, type_hash
+
         fetcher = OperationFetcher(
             datasource=self._datasource,
             first_level=first_level,
@@ -155,31 +189,35 @@ class OperationIndex(Index):
             transaction_addresses=transaction_addresses,
             origination_addresses=origination_addresses,
             cache=cache,
+            migration_originations=migration_originations,
         )
 
         async for level, operations in fetcher.fetch_operations_by_level():
             await self._process_level_operations(level, operations)
 
-        state.level = last_level  # type: ignore
-        await state.save()
+        await self._exit_sync_state(last_level)
 
     async def _process_level_operations(self, level: int, operations: List[OperationData], block: Optional[HeadBlockData] = None) -> None:
-        state = await self.get_state()
-        if level < state.level:
-            raise RuntimeError(f'Level of operation batch is lower than index state level: {level} < {state.level}')
+        if level < self.state.level:
+            raise RuntimeError(f'Level of operation batch is lower than index state level: {level} < {self.state.level}')
 
         if self._rollback_level:
-            if state.level != self._rollback_level:
-                raise RuntimeError(f'Rolling back to level {self._rollback_level}, state level {state.level}')
-            if level != self._rollback_level:
-                raise RuntimeError(f'Rolling back to level {self._rollback_level}, got operations of level {level}')
+            levels = {
+                'operations': level,
+                'rollback': self._rollback_level,
+                'index': self.state.level,
+                'block': block.level if block else level,
+            }
+            if len(set(levels.values())) != 1:
+                levels_repr = ', '.join(f'{k}={v}' for k, v in levels.items())
+                raise RuntimeError(f'Index is in a rollback state, but received operation batch with different levels: {levels_repr}')
 
             self._logger.info('Rolling back to previous level, verifying processed operations')
             expected_hashes = set(self._last_hashes)
             received_hashes = set([op.hash for op in operations])
             reused_hashes = received_hashes & expected_hashes
             if reused_hashes != expected_hashes:
-                self._logger.warning('Attempted a single level rollback but arrived block differs from processed one')
+                self._logger.warning('Attempted a single level rollback, but arrived block differs from processed one')
                 await self._ctx.reindex()
 
             self._rollback_level = None
@@ -193,10 +231,10 @@ class OperationIndex(Index):
             self._logger.info('Processing %s operations of level %s', len(operations), level)
             await self._process_operations(operations)
 
-            state.level = level  # type: ignore
+            self.state.level = level  # type: ignore
             if block:
-                state.hash = block.hash  # type: ignore
-            await state.save()
+                self.state.hash = block.hash  # type: ignore
+            await self.state.save()
 
     async def _match_operation(self, pattern_config: OperationHandlerPatternConfigT, operation: OperationData) -> bool:
         """Match single operation with pattern"""
@@ -272,14 +310,12 @@ class OperationIndex(Index):
                         operation_idx += 1
 
                     if pattern_idx == len(handler_config.pattern):
-                        self._logger.info('%s: `%s` handler matched!', operation_subgroup.hash, handler_config.callback)
                         await self._on_match(operation_subgroup, handler_config, matched_operations)
 
                         matched_operations = []
                         pattern_idx = 0
 
                 if len(matched_operations) >= sum(map(lambda x: 0 if x.optional else 1, handler_config.pattern)):
-                    self._logger.info('%s: `%s` handler matched!', operation_subgroup.hash, handler_config.callback)
                     await self._on_match(operation_subgroup, handler_config, matched_operations)
 
     async def _on_match(
@@ -289,6 +325,9 @@ class OperationIndex(Index):
         matched_operations: List[Optional[OperationData]],
     ):
         """Prepare handler arguments, parse parameter and storage. Schedule callback in executor."""
+        self._logger.info('%s: `%s` handler matched!', operation_subgroup.hash, handler_config.callback)
+        start = time.perf_counter()
+
         args: List[Optional[Union[Transaction, Origination, OperationData]]] = []
         for pattern_config, operation in zip(handler_config.pattern, matched_operations):
             if operation is None:
@@ -343,10 +382,14 @@ class OperationIndex(Index):
             logger=logger,
             template_values=self._config.template_values,
             datasource=self.datasource,
+            index_config=self._config,
         )
 
         await handler_config.callback_fn(handler_context, *args)
 
+        end = time.perf_counter()
+        self._logger.debug(f'{handler_config.callback}` handler executed in {end - start:.3f}s')
+
         if handler_context.updated:
             self._ctx.commit()
 
@@ -375,7 +418,7 @@ class OperationIndex(Index):
                             addresses.add(address)
         return addresses
 
-    async def _get_contract_hashes(self, address: str) -> Tuple[str, str]:
+    async def _get_contract_hashes(self, address: str) -> Tuple[int, int]:
         if address not in self._contract_hashes:
             summary = await self._datasource.get_contract_summary(address)
             self._contract_hashes[address] = (summary['codeHash'], summary['typeHash'])
@@ -387,10 +430,10 @@ class BigMapIndex(Index):
 
     def __init__(self, ctx: DipDupContext, config: BigMapIndexConfig, datasource: TzktDatasource) -> None:
         super().__init__(ctx, config, datasource)
-        self._queue: Deque[Tuple[int, List[BigMapData]]] = deque()
+        self._queue: Deque[Tuple[int, List[BigMapData], Optional[HeadBlockData]]] = deque()
 
-    def push(self, level: int, big_maps: List[BigMapData]):
-        self._queue.append((level, big_maps))
+    def push(self, level: int, big_maps: List[BigMapData], block: Optional[HeadBlockData] = None):
+        self._queue.append((level, big_maps, block))
 
     async def _process_queue(self):
         if not self._queue:
@@ -398,19 +441,16 @@ class BigMapIndex(Index):
         self._logger.info('Processing websocket queue')
         with suppress(IndexError):
             while True:
-                level, big_maps = self._queue.popleft()
-                await self._process_level_big_maps(level, big_maps)
+                level, big_maps, block = self._queue.popleft()
+                await self._process_level_big_maps(level, big_maps, block)
 
     async def _synchronize(self, last_level: int, cache: bool = False) -> None:
         """Fetch operations via Fetcher and pass to message callback"""
-        state = await self.get_state()
-        first_level = state.level
-        if first_level == last_level:
+        first_level = await self._enter_sync_state(last_level)
+        if first_level is None:
             return
-        if first_level > last_level:
-            raise RuntimeError(first_level, last_level)
 
-        self._logger.info('Fetching operations from level %s to %s', first_level, last_level)
+        self._logger.info('Fetching big map diffs from level %s to %s', first_level, last_level)
 
         big_map_addresses = await self._get_big_map_addresses()
         big_map_paths = await self._get_big_map_paths()
@@ -427,20 +467,20 @@ class BigMapIndex(Index):
         async for level, big_maps in fetcher.fetch_big_maps_by_level():
             await self._process_level_big_maps(level, big_maps)
 
-        state.level = last_level  # type: ignore
-        await state.save()
+        await self._exit_sync_state(last_level)
 
-    async def _process_level_big_maps(self, level: int, big_maps: List[BigMapData]):
-        state = await self.get_state()
-        if state.level >= level:
-            raise RuntimeError(state.level, level)
+    async def _process_level_big_maps(self, level: int, big_maps: List[BigMapData], block: Optional[HeadBlockData] = None):
+        if level < self.state.level:
+            raise RuntimeError(f'Level of operation batch is lower than index state level: {level} < {self.state.level}')
 
         async with in_global_transaction():
             self._logger.info('Processing %s big map diffs of level %s', len(big_maps), level)
             await self._process_big_maps(big_maps)
 
-            state.level = level  # type: ignore
-            await state.save()
+            self.state.level = level  # type: ignore
+            if block:
+                self.state.hash = block.hash  # type: ignore
+            await self.state.save()
 
     async def _match_big_map(self, handler_config: BigMapHandlerConfig, big_map: BigMapData) -> bool:
         """Match single big map diff with pattern"""
@@ -456,6 +496,8 @@ class BigMapIndex(Index):
         matched_big_map: BigMapData,
     ) -> None:
         """Prepare handler arguments, parse key and value. Schedule callback in executor."""
+        self._logger.info('%s: `%s` handler matched!', matched_big_map.operation_id, handler_config.callback)
+        start = time.perf_counter()
 
         if matched_big_map.action.has_key:
             key_type = handler_config.key_type_cls
@@ -492,10 +534,14 @@ class BigMapIndex(Index):
             logger=logger,
             template_values=self._config.template_values,
             datasource=self.datasource,
+            index_config=self._config,
         )
 
         await handler_config.callback_fn(handler_context, big_map_context)
 
+        end = time.perf_counter()
+        self._logger.debug(f'{handler_config.callback}` handler executed in {end - start:.3f}s')
+
         if handler_context.updated:
             self._ctx.commit()
 
@@ -506,7 +552,6 @@ class BigMapIndex(Index):
             for handler_config in self._config.handlers:
                 big_map_matched = await self._match_big_map(handler_config, big_map)
                 if big_map_matched:
-                    self._logger.info('%s: `%s` handler matched!', big_map.operation_id, handler_config.callback)
                     await self._on_match(handler_config, big_map)
 
     async def _get_big_map_addresses(self) -> Set[str]:
